lappend auto_path [file join [pwd] lib]

package require Tk 8.6
package require Ttk 8.6
package require cmdbar
package require config 
package require macro
package require audio
package require TkEsweepXYPlot
package require overlay

# source basic modules
source [file join [pwd] lib config.tcl]
#source [file join [pwd] lib cmds.tcl]
source [file join [pwd] lib gui.tcl]
source [file join [pwd] lib file.tcl]
source [file join [pwd] lib primes.tcl]

####################
# global variables #
####################

# ID of master curve
set ::master 1
trace add variable ::master write [list apply {{v1 v2 op} {
    upvar $v1 master
    ::overlay::removeTag all master
    ::overlay::addTag $master master -background orange
    :cursor trace 1 #$master
  }}]

proc IMDupdateFunc {updArgs id _overlays prop op} {
  # since this function is called by a trace, we need to upvar 2 levels
  upvar 2 ${_overlays} overlays
  switch -- $op {
    unset {
      catch {::TkEsweepXYPlot::removeTrace IMD $id}
      catch {::TkEsweepXYPlot::removeTrace IMD ${id}_IMD}
      ::TkEsweepXYPlot::plot IMD
    }
    default {
      switch -- $prop {
        Show {
          if {[::TkEsweepXYPlot::exists IMD $id]} { 
            if {$overlays($id,Show) eq {no}} {
              ::TkEsweepXYPlot::configTrace IMD $id -state off
              ::TkEsweepXYPlot::configTrace IMD ${id}_IMD -state off
            } else {
              ::TkEsweepXYPlot::configTrace IMD $id -state on
              ::TkEsweepXYPlot::configTrace IMD ${id}_IMD -state on
            }
            ::TkEsweepXYPlot::plot IMD
          }
          # else: 
          # this case most probably happens when a trace is added, 
          # and the properties are not set in the correct order (1st: FR, then: the rest)
        }
        FR -
        default {
          showIMD $id
        }
      }
    }
  }
}

# register the update function
set ::overlay::updateProps [list Name FR Color Show]
set ::overlay::updateFunc [list IMDupdateFunc {_}]


######################
# Overlay management #
######################

proc initOverlays {w} {
  ::overlay::addProperty Color      {}        0  {Display color}
  ::overlay::addProperty Name       {}        200 {Descriptive name}
  ::overlay::addProperty FR         {}       0  {Measured frequency response}
  ::overlay::addProperty IMD         {}       0  {Measured distortion response}
  ::overlay::addProperty Show       {yes}     0  {Show or hide this overlay (boolean)}
  ::overlay::addProperty OutLvl     {}    0  {Output level (in V)}
  ::overlay::addProperty InLvl     {}    0  {Input level (in Vrms)}

  set win [::overlay::init $w]

  pack $win -side right -expand yes -fill both
}

proc :delete {ID} {
  if {$ID ne {all}} {
    if {![regexp {(^#)([[:digit:]]+$)} $ID -> hash id]} {
        return -code error "Parameter not an ID: $ID"
    }
  } else {
    set id $ID
  }
  ::overlay::delete $id
  ::TkEsweepXYPlot::removeTrace IMP $id
  catch {updateIMP}
}

##################
# Macro handling #
##################

proc init_macros {} {
  load_macro
  ::macro::addAlias :w :l :set :measure :delete :expr :show :hide :q :rename :modify :message :copy :paste :query 
  if {$::Debug} {
    ::macro::addAlias puts
  }
}

proc :measure {{mic default}} {
  global master
  set FS [setup cget Audio,Samplerate]

  # create the signal
  lassign [setup cget Signal,Definition,Bandwidth] f0 f1
  set N [setup cget Signal,Definition,Bins]
  set spec [setup cget Signal,Definition,Spectrum]
  set ph [setup cget Signal,Definition,Phase]
  set spacing [setup cget Signal,Definition,Spacing]

  switch -- $spacing {
    lin -
    log -
    default {
      set step [getStep $f0 $f1 $N $spacing]

      # the size of the signal depends on the smallest frequency fraction
      # this can be either the lowest frequency (f0) or the difference between the first two bins (df)
      set df [expr ($f0 $step)-$f0]
      if {$f0 < $df} {
        set size [expr {1.0*$FS/$f0}]
      } else {
        set size [expr {1.0*$FS/$df}]
      }
      if {$size < [setup cget Signal,MinSize]} {
        set size [setup cget Signal,MinSize]
      }
      # make it a power of two
      for {set n 1} {$n < $size} {set n [expr {$n << 1}]} {}
      set df [expr {double($FS)/$n}]

      # create two signals: one keeps the sum of all signals, in another a new sinewave will be created 
      set ref [esweep::create -type wave -samplerate $FS -size $n]
      set tmp [esweep::create -type wave -samplerate $FS -size $n]

      for {set i 0; set f $f0} {$i < $N} {incr i} {
        if {$ph eq {random}} {
          set phase [expr {rand()*acos(-1)}]
        } elseif {[llength $ph] == 1} {
          set phase $ph
        } else {
          set phase [lindex $ph $i]
        }
        set k [esweep::generate sine -obj tmp -frequency $f -phase $phase]
        esweep::expr ref + $tmp
        # append the real generated frequency to the signal definition
        lappend sigdef $k [expr {$k*$df}] $phase
        # but keep the original frequency for correct spacing
        set f [expr $f $step]
      }
    }
  }
  set max [esweep::max -obj $ref]
  set min [esweep::min -obj $ref]
  set peak [expr {abs($max > $min ? $max : $min)}]
  set rms [expr {[esweep::sqsum -obj $ref]/$n}]
  set crest [expr {$peak/$rms}]

  # compute output level
  set ch_out [setup cget Output,Channel]
  if {[catch {setup cget Output,$ch_out,Gain}] || [catch {setup cget Output,$ch_out,Sensitivity}]} {
    set ch Global
    ::cmdbar::writeToCmdBar \
      "Output channel $ch_out not fully specified, using global output parameters" -bg yellow
  }
  set outsense [setup cget Output,$ch_out,Sensitivity]
  set outgain [setup cget Output,$ch_out,Gain]
  set lvl [expr {$crest*[setup cget Output,Level] / ($outgain * $outsense)}]
  if {$lvl > [setup cget Output,MaxFS]} {
    set lvl [setup cget Output,MaxFS]
    setup configure Output,Level [format "%.3f" [expr \
      {$lvl*[setup cget Output,$ch_out,Gain]*[setup cget Output,$ch_out,Sensitivity]/$crest}]]
    ::cmdbar::writeToCmdBar \
      "Output level exceeded maximum allowed output. Adjusted to [setup cget Output,Level] V" -bg yellow
  }
  set lvl [expr {$lvl/$peak}]
  esweep::expr ref * $lvl

  # the following steps are necessary to make the audio I/O FFT-ready
  # 1. we need a fixed framesize in the audio device
  audio configure Framesize [set framesize 2048]
  # 2. the total length must be a multiple of the framesize
  set record_length [expr {1000.0*$n/$FS+[setup cget Audio,Latency]}]
  set rn [expr {int(0.5+$FS*$record_length/1000.0)}]
  set rn [expr {$rn+$framesize-$rn%$framesize}]
  set record_length [expr {1000.0*$rn/$FS}]
  # 3. copy the reference data to a new output object
  set out [esweep::create -type [esweep::type -obj $ref] -size $rn -samplerate $FS]
  esweep::copy -dst out -src $ref; # copy first part
  esweep::copy -dst out -src $ref -dstIdx $n; # fill the output completely
  # now we have a signal with a period of $n samples
  # with total length $rn

  # we can measure always in the dual channel mode
  lassign [::audio::measure dual \
            -ref out \
            -input [list [setup cget Input,Channel] [setup cget Input,Reference]] \
            -output [setup cget Output,Channel] \
            -flags {rec_only} \
            -record $record_length] in inref
  
  if {[catch {esweep::size -obj $in}]} {
    ::cmdbar::writeToCmdBar {Measurement cancelled} -bg yellow
    return
  }

  # we need the second halves of the recorded signals
  set in [esweep::get -obj $in -from [expr {$rn-$n}] -to [expr {$rn-1}]]
  set inref [esweep::get -obj $inref -from [expr {$rn-$n}] -to [expr {$rn-1}]]
  if {[setup cget Measure,Mode] eq {dual}} {
    # get the real output level
    set ch_ref [setup cget Input,Reference]
    if {[catch {setup cget Input,$ch_ref,Gain}] || [catch {setup cget Input,$ch_ref,Sensitivity}]} {
      set ch_ref Global
      ::cmdbar::writeToCmdBar \
        "Reference channel $ch_ref not fully specified, using global input parameters" -bg yellow
    }
    set refsense [setup cget Input,$ch_ref,Sensitivity]
    set refgain [setup cget Input,$ch_ref,Gain]
    set sc [expr {$refsense/$refgain}]

    set outlevel [expr {$sc*sqrt([esweep::sqsum -obj $inref]/[esweep::size -obj $inref])}]
  } else {
    set outlevel [setup cget Output,Level]
  }

  # scale input signal
  set ch_in [setup cget Input,Channel]
  if {[catch {setup cget Input,$ch_in,Gain}] || [catch {setup cget Input,$ch_in,Sensitivity}]} {
    set ch_in Global
    ::cmdbar::writeToCmdBar \
      "Input channel $ch_in not fully specified, using global input parameters" -bg yellow
  }
  set insense [setup cget Input,$ch_in,Sensitivity]
  set ingain [setup cget Input,$ch_in,Gain]
  if {$mic eq {default}} {set mic [setup cget Mic,Default]}
  set micsense [setup cget Mic,$mic,Sensitivity]
  set micgain [setup cget Mic,$mic,Gain]

  set sc [expr {1000*$insense/($micsense*$ingain*$micgain*2e-5)}]
  esweep::expr in * $sc
  set rms [expr {sqrt([esweep::sqsum -obj $in]/[esweep::size -obj $in])}]

  esweep::fft -inplace -obj in
  esweep::toPolar -obj in
  # we need to scale by the FFT size because this FFT does not scale by 1/(N/2) by itself
  esweep::expr in / [expr {[esweep::size -obj $in]/2}]
   
  # get the total power in each bandwidth
  lassign [esweep::index -index 1 -obj $in] df
  lassign [setup cget Processing,Bandwidth] f0 f1
  set N [setup cget Processing,Bins]

  #set rms 0.0
  # on older architectures the esweep::setIndex function is not available
  # we test for it and use an alternative processing method if applicable
  if {[catch {esweep::setIndex -index 0 -obj in [list 0.0 0.0]}]} {
    # extract the frequency bins
    foreach {i f p} $sigdef {
      # get the value at frequency $f 
      lassign [esweep::index -index $i -obj $in] f abs arg
      #set rms [expr {$rms+$abs**2}]

      # FR trace format: 1st element: frequency, 2nd element: magnitude, 3rd element: arg
      lappend fr $f [expr {20*log10($abs)}] $arg
    }

    switch -- [setup cget Processing,Spacing] {
      lin {
        set step [getStep $f0 $f1 $N lin]
        # lower bw limit
        set fl [expr $f0-1.0/(2*(0 $step))]
        # upper bw limit
        set fo [expr $fl $step]
      }
      log -
      default {
        set step [getStep $f0 $f1 $N $spacing]
        # lower bw limit
        set fl [expr $f0/sqrt(1 $step)]
        # upper bw limit
        set fo [expr $fl $step]
      }
    }
    for {set k 0; set f $f0} {$k < $N} {incr k} {
      # get distortion in frequency band
      set n 0
      set pw 0
      for {set j [set j0 [expr {int(0.5+$fl/$df)}]]} {$j < [set j1 [expr {int(0.5+$fo/$df)}]]} {incr j} {
        if {$j < 0} {continue}
        if {$j >= [esweep::size -obj $in]} {break}
        lassign [esweep::index -index $j -obj $in] fimd d
        # total power in band
        set pw [expr {$pw+$d**2}]
        incr n
      }
      # subtract the level of the main peak in this bandwidth
      foreach {i fi p} $sigdef {
        if {$i < $j0} {
          continue
        } elseif {$i >= $j1} {
          break
        } else {
          lassign [esweep::index -index $i -obj $in] fk d
          set pw [expr {$pw-$d**2}]
          incr n -1
        }
      }

      if {$n > 0} {lappend imd $f [expr {10*log10($pw/$n)}] 0.0}
      set f [expr $f $step]
      set fl $fo
      set fo [expr $fo $step]
    }
  } else {
    foreach {i f p} $sigdef {
      # get the value at frequency $f 
      lassign [esweep::index -index $i -obj $in] f abs arg
      # set this bin to 0
      esweep::setIndex -obj in -index $i -value [list 0.0 $arg]
      #set rms [expr {$rms+$abs**2}]

      # FR trace format: 1st element: frequency, 2nd element: magnitude, 3rd element: arg
      lappend fr $f [expr {20*log10($abs)}] $arg
    }

    set smfac [expr {$N*log10(2)/log10(double($f1)/$f0)}]
    esweep::smooth -obj in -factor $smfac
    esweep::lg -obj in
    esweep::expr in * 20
    set imd [getLogSpacedFr $in [list $f0 $f1 $N]]
  }

  set props [list Name Master \
                  Color red \
                  FR  $fr \
                  IMD  $imd \
                  OutLvl $outlevel \
                  InLvl  $rms \
                  Show {yes}]

  if {[::overlay::exists $::master]} {
    ::overlay::modify $::master $props
  } else {
    set ::master [::overlay::add $props]
  }
  ::overlay::select set $::master
  # copy master to buffer
  ::overlay::copy
  #showIMD $::master
  return #$::master
}

proc :expr {ID op IDorExpr} {
  if {![regexp {^[-/\+\*]{1}$} $op]} {
    ::cmdbar::writeToCmdBar \
      return -code error "Invalid operator $op"
  }
  if {![regexp {(^#)([[:digit:]]+$)} $ID -> hash ID1]} {
    ::cmdbar::writeToCmdBar \
      return -code error "First operand not an ID: $ID"
  } else {
    set ir1 [esweep::clone -src [getProp $ID1 IMP]]
  }
  if {[regexp {(^#)([[:digit:]]+$)} $IDorExpr -> hash ID2]} {
    # is an ID
    set ir2 [esweep::clone -src [getProp $ID2 IMP]]
  } elseif {[catch {set ir2 [expr $IDorExpr]} msg]} {
    ::cmdbar::writeToCmdBar \
      return -code error "Second operand neither an ID nor an expression: $IDorExpr"
  }
  if {[catch {esweep::expr ir1 $op $ir2} msg]} {
    ::cmdbar::writeToCmdBar \
      return -code error "Expression failed with: $msg"
  }
  ::overlay::modify $ID1 [list IMP $ir1]
  #showIMP $ID1
}

###########
# Startup #
###########
default_config
init_macros
appWindows
bindings .
load_config
